import re
import attr
from enum import Enum
from alita.base import BaseRoute, BaseRouter
from alita.converters import CONVERTER_TYPES
from urllib.parse import urljoin, quote
from alita.helpers import get_request_url, set_query_parameter
from alita.exceptions import NotFound, BadRequest, RequestSlash, RequestRedirect
from alita.response import RedirectResponse


class NoMatchFound(NotFound):
    pass


class Match(Enum):
    NONE = 0
    PARTIAL = 1
    FULL = 2


def replace_params(path, param_converters, path_params):
    for key, value in list(path_params.items()):
        if "{" + key + "}" in path:
            converter = param_converters[key]
            value = converter.to_string(value)
            path = path.replace("{" + key + "}", value)
            path_params.pop(key)
    for key, value in path_params.items():
        path = set_query_parameter(path, key, value)
    return path


# Match parameters in URL paths, eg. '<param_name:int/float/str/path>'
PARAM_REGEX = re.compile("<([a-zA-Z_][a-zA-Z0-9_]*)(:[a-zA-Z_][a-zA-Z0-9_]*)?>")


class URLPath:
    def __init__(self, path, protocol="http", host="", params=None):
        self.path = path
        self.protocol = protocol
        self.host = host
        self.params = params or {}
        assert protocol in ("http", "websocket")

    def make_url(self, base_url='/', request=None):
        if request is not None:
            assert not base_url, 'Cannot set both "base_url" and "request".'
            return get_request_url(request)
        elif base_url:
            assert not request, 'Cannot set both "request" and "**components".'
            return urljoin(base_url, self.path)
        else:
            raise ValueError('base_url or request need supply one.')


@attr.s
class RouteMatch:
    status = attr.ib(validator=attr.validators.instance_of(Match))
    endpoint = attr.ib(validator=attr.validators.instance_of(str))
    view_func = attr.ib()
    path_params = attr.ib(validator=attr.validators.instance_of(dict))


class Route(BaseRoute):
    def __init__(self, path, endpoint, view_func, methods=None, strict_slashes=None):
        assert path.startswith("/"), "Routed paths must start with '/'"
        self.path = path
        self.endpoint = endpoint
        self.view_func = view_func
        self.strict_slashes = strict_slashes
        self.is_leaf = not self.path.endswith('/')
        methods = methods or getattr(view_func, 'methods', None) or ('GET',)
        self.methods = set([method.upper() for method in methods])
        if "GET" in self.methods:
            self.methods |= set(["HEAD"])
        self.rule = self.is_leaf and self.path or self.path.rstrip('/')
        self.path_regex, self.path_format, self.param_converters = self.compile_path()

    def compile_path(self):
        path_regex = ""
        path_format = ""

        idx = 0
        param_converters = {}
        for match in PARAM_REGEX.finditer(self.rule):
            param_name, converter_type = match.groups("str")
            converter_type = converter_type.lstrip(":")
            if converter_type not in CONVERTER_TYPES:
                raise RuntimeError(f"Unknown path converter '{converter_type}'")
            converter = CONVERTER_TYPES[converter_type]
            path_regex += self.rule[idx: match.start()]
            path_regex += f"(?P<{param_name}>{converter.regex})"
            path_format += self.rule[idx: match.start()]
            path_format += "{%s}" % param_name
            param_converters[param_name] = converter
            idx = match.end()

        path_regex += self.rule[idx:]
        path_format += self.rule[idx:]
        regex = r'^%s%s$' % (
            path_regex,
            (not self.is_leaf or not self.strict_slashes) and
            '(?<!/)(?P<__suffix__>/?)' or ''
        )
        return re.compile(regex), path_format, param_converters

    def match(self, request):
        if request.scheme in ("http", "https", "wss", "ws"):
            match = self.path_regex.search(request.path)
            if not match:
                raise NoMatchFound()
            matched_params = match.groupdict()
            if self.strict_slashes and not self.is_leaf and \
                    not matched_params.pop('__suffix__'):
                raise RequestSlash()
            for key, value in matched_params.items():
                matched_params[key] = self.param_converters[key].convert(value)
            if self.methods and request.method not in self.methods:
                status = Match.PARTIAL
            else:
                status = Match.FULL
            return RouteMatch(
                status,
                self.endpoint,
                self.view_func,
                matched_params,
            )
        else:
            raise NoMatchFound()

    def url_path_for(self, endpoint, **path_params):
        seen_keys = set(path_params.keys())
        expected_keys = set(self.param_converters.keys())
        if endpoint != self.endpoint:
            raise NoMatchFound()
        if expected_keys and not seen_keys.issuperset(expected_keys):
            raise NoMatchFound()
        path = replace_params(self.path_format, self.param_converters, path_params)
        return URLPath(path=path, protocol="http", params=path_params)

    def __eq__(self, other):
        return (
            isinstance(other, Route)
            and self.path == other.path
            and self.endpoint == other.endpoint
            and self.methods == other.methods
        )


class Router(BaseRouter):
    def __init__(self, app, routes=None):
        self.app = app
        self.routes = routes or []

    def add_route(self, path, endpoint, view_func, methods=None, strict_slashes=None):
        if strict_slashes is None:
            strict_slashes = self.app.config.get('STRICT_SLASHES', True)
        route = Route(
            path,
            endpoint=endpoint,
            view_func=view_func,
            methods=methods,
            strict_slashes=strict_slashes
        )
        self.routes.append(route)

    def match(self, request):
        for route in self.routes:
            try:
                route_math = route.match(request)
                if route_math.status == Match.PARTIAL:
                    raise BadRequest()
                return route_math
            except NoMatchFound:
                pass
            except RequestSlash:
                new_path = quote(request.path, safe='/:|+') + '/'
                raise RequestRedirect(response=RedirectResponse(
                    get_request_url(request, path=new_path)))
        raise NoMatchFound()

    def url_path_for(self, endpoint, **path_params):
        for route in self.routes:
            try:
                return route.url_path_for(endpoint, **path_params)
            except NoMatchFound:
                pass
        raise NoMatchFound()



import os
import sys
import importlib
from alita.base import BaseFactory


def prepare_import(path):
    """
    Given a filename this will try to calculate the python path, add it
    to the search path and return the actual module name that is expected.
    """
    path = os.path.realpath(path)

    if os.path.splitext(path)[1] == '.py':
        path = os.path.splitext(path)[0]

    if os.path.basename(path) == '__init__':
        path = os.path.dirname(path)

    module_name = []

    # move up until outside package structure (no __init__.py)
    while True:
        path, name = os.path.split(path)
        module_name.append(name)

        if not os.path.exists(os.path.join(path, '__init__.py')):
            break

    if sys.path[0] != path:
        sys.path.insert(0, path)

    return '.'.join(module_name[::-1])


class CliFactory:
    def load_app(self, app_file=None):
        try:
            if not app_file:
                app_file = os.environ.get('ALITA_APP') or "app.py"
            import_name = prepare_import(app_file)
            app_module = importlib.import_module(import_name)
            return getattr(app_module, 'app')
        except SyntaxError as e:
            message = (
                          'Unable to import your app.py file:\n\n'
                          'File "%s", line %s\n'
                          '  %s\n'
                          'SyntaxError: %s'
                      ) % (getattr(e, 'filename'), e.lineno, e.text, e.msg)
            raise RuntimeError(message)


class AppFactory(BaseFactory):
    def __init__(self, app=None):
        self.app = app

    def create_request_object(self, environ):
        from alita.request import Request
        return Request(self.app, environ)

    def create_base_response_class(self):
        from alita.response import HTTPResponse
        return HTTPResponse

    def create_base_exception_class(self):
        from alita.exceptions import HTTPException
        return HTTPException

    def create_exception_handler_object(self):
        from alita.handler import ExceptionHandler
        return ExceptionHandler(self.app)

    def create_router_object(self):
        from alita.routing import Router
        return Router(self.app)

    def create_static_handler(self):
        from alita.handler import StaticHandler
        return StaticHandler(self.app)

    def create_jinja_loader(self):
        from alita.templating import DispatchingJinjaLoader
        return DispatchingJinjaLoader(self.app)



# -*- coding: utf-8 -*-
"""
alita template.
"""
from alita.response import HtmlResponse
from jinja2 import BaseLoader, Environment as BaseEnvironment, \
     TemplateNotFound
from alita.signals import template_rendered, before_render_template


class Environment(BaseEnvironment):
    """Works like a regular Jinja2 environment but has some additional
    knowledge of how Flask's blueprint works so that it can prepend the
    name of the blueprint to referenced templates if necessary.
    """

    def __init__(self, app, **options):
        if 'loader' not in options:
            options['loader'] = app.create_global_jinja_loader()
        BaseEnvironment.__init__(self, **options)
        self.app = app


class DispatchingJinjaLoader(BaseLoader):
    """
    A loader that looks for templates in the application and all
    the blueprint folders.
    """

    def __init__(self, app):
        self.app = app

    def get_source(self, environment, template):
        for srcobj, loader in self._iter_loaders(template):
            try:
                return loader.get_source(environment, template)
            except TemplateNotFound:
                continue
        raise TemplateNotFound(template)

    def _iter_loaders(self, template):
        loader = self.app.jinja_loader
        if loader is not None:
            yield self.app, loader

        for blueprint in self.app.iter_blueprints():
            loader = blueprint.jinja_loader
            if loader is not None:
                yield blueprint, loader

    def list_templates(self):
        result = set()
        loader = self.app.jinja_loader
        if loader is not None:
            result.update(loader.list_templates())

        for blueprint in self.app.iter_blueprints():
            loader = blueprint.jinja_loader
            if loader is not None:
                for template in loader.list_templates():
                    result.add(template)

        return list(result)


async def _render(request, context, template):
    """Renders the template and fires the signal"""
    await request.update_template_context(context)
    before_render_template.send(request.app, template=template, context=context)
    rv = await template.render_async(context)
    template_rendered.send(request.app, template=template, context=context)
    return rv


async def render_template(request, template_name_or_list, **context):
    """
    Renders a template from the template folder with the given
    context.

    :param request: app request object.
    :param template_name_or_list: the name of the template to be
                                  rendered, or an iterable with template names
                                  the first one existing will be rendered
    :param context: the variables that should be available in the
                    context of the template.
    """
    return HtmlResponse(await _render(
        request, context,
        request.app.jinja_env.get_or_select_template(template_name_or_list)))


async def render_template_string(request, source, **context):
    """
    Renders a template from the given template source string
    with the given context. Template variables will be autoescaped.

    :param request: app request object.
    :param source: the source code of the template to be
                   rendered
    :param context: the variables that should be available in the
                    context of the template.
    """
    return HtmlResponse(await _render(
        request, context, request.app.jinja_env.from_string(source)))



import os
import sys
import time
import signal
import logging
import subprocess
from itertools import chain
from multiprocessing import Process


def _iter_module_files():
    """This iterates over all relevant Python files.  It goes through all
    loaded files from modules, all files in folders of already loaded modules
    as well as all files reachable through a package.
    """
    # The list call is necessary on Python 3 in case the module
    # dictionary modifies during iteration.
    for module in list(sys.modules.values()):
        if module is None:
            continue
        filename = getattr(module, '__file__', None)
        if filename:
            if os.path.isdir(filename) and \
               os.path.exists(os.path.join(filename, "__init__.py")):
                filename = os.path.join(filename, "__init__.py")

            old = None
            while not os.path.isfile(filename):
                old = filename
                filename = os.path.dirname(filename)
                if filename == old:
                    break
            else:
                if filename[-4:] in ('.pyc', '.pyo'):
                    filename = filename[:-1]
                yield filename


def _get_args_for_reloading():
    """Returns the executable. This contains a workaround for windows
    if the executable is incorrectly reported to not have the .exe
    extension which can cause bugs on reloading.
    """
    rv = [sys.executable]
    py_script = sys.argv[0]
    if os.name == 'nt' and not os.path.exists(py_script) and \
       os.path.exists(py_script + '.exe'):
        py_script += '.exe'
    if os.path.splitext(rv[0])[1] == '.exe' and os.path.splitext(py_script)[1] == '.exe':
        rv.pop(0)
    rv.append(py_script)
    rv.extend(sys.argv[1:])
    return rv


def kill_process_children_unix(pid):
    """Find and kill child processes of a process (maximum two level).

    :param pid: PID of parent process (process ID)
    :return: Nothing
    """
    root_process_path = "/proc/{pid}/task/{pid}/children".format(pid=pid)
    if not os.path.isfile(root_process_path):
        return
    with open(root_process_path) as children_list_file:
        children_list_pid = children_list_file.read().split()

    for child_pid in children_list_pid:
        children_proc_path = "/proc/%s/task/%s/children" % (
            child_pid,
            child_pid,
        )
        if not os.path.isfile(children_proc_path):
            continue
        with open(children_proc_path) as children_list_file_2:
            children_list_pid_2 = children_list_file_2.read().split()
        for _pid in children_list_pid_2:
            try:
                os.kill(int(_pid), signal.SIGTERM)
            except ProcessLookupError:
                continue
        try:
            os.kill(int(child_pid), signal.SIGTERM)
        except ProcessLookupError:
            continue


def kill_process_children_osx(pid):
    """Find and kill child processes of a process.

    :param pid: PID of parent process (process ID)
    :return: Nothing
    """
    subprocess.run(["pkill", "-P", str(pid)])


def kill_process_children(pid):
    """Find and kill child processes of a process.

    :param pid: PID of parent process (process ID)
    :return: Nothing
    """
    if sys.platform == "darwin":
        kill_process_children_osx(pid)
    elif sys.platform == "linux":
        kill_process_children_unix(pid)
    else:
        pass  # should signal error here


def kill_program_completly(proc):
    """Kill worker and it's child processes and exit.

    :param proc: worker process (process ID)
    :return: Nothing
    """
    kill_process_children(proc.pid)
    proc.terminate()
    os._exit(0)


class ReloadLoop(object):
    name = None
    _sleep = staticmethod(time.sleep)

    def __init__(self, extra_files=None, interval=1, _log=None):
        self.extra_files = set(os.path.abspath(x)
                               for x in extra_files or ())
        self.interval = interval
        self._log = _log or self._loop_log
        self.logger = logging.getLogger(__name__)

    def _loop_log(self, type, message, *args, **kwargs):
        getattr(self.logger, type)(message.rstrip(), *args, **kwargs)

    def run(self):
        pass

    def restart_with_reloader(self):
        """Spawn a new Python interpreter with the same arguments as this one,
        but running the reloader thread.
        """
        while True:
            self._log('info', ' * Restarting with %s' % self.name)
            args = _get_args_for_reloading()
            new_environ = os.environ.copy()
            new_environ['SERVER_RUN_MAIN'] = 'true'
            cmd = " ".join(args)
            worker_process = Process(
                target=subprocess.call,
                args=(cmd,),
                kwargs={
                    "cwd": os.getcwd(),
                    "shell": True,
                    "env": new_environ
                },
            )
            worker_process.start()
            return worker_process

    def trigger_reload(self, filename):
        self.log_reload(filename)
        sys.exit(3)

    def log_reload(self, filename):
        filename = os.path.abspath(filename)
        self._log('info', ' * Detected change in %r, reloading' % filename)


class AutoReloadLoop(ReloadLoop):
    name = 'stat'

    def run(self):
        mtimes = {}
        worker_process = self.restart_with_reloader()
        signal.signal(
            signal.SIGTERM, lambda *args: kill_program_completly(worker_process)
        )
        signal.signal(
            signal.SIGINT, lambda *args: kill_program_completly(worker_process)
        )
        while True:
            for filename in chain(_iter_module_files(),
                                  self.extra_files):
                try:
                    mtime = os.stat(filename).st_mtime
                except OSError:
                    continue

                old_time = mtimes.get(filename)
                if old_time is None:
                    mtimes[filename] = mtime
                    continue
                elif mtime > old_time:
                    kill_process_children(worker_process.pid)
                    worker_process.terminate()
                    worker_process = self.restart_with_reloader()
                    mtimes[filename] = mtime
                    break
            self._sleep(self.interval)


def run_auto_reload(extra_files=None, interval=1, _log=None):
    try:
        AutoReloadLoop(extra_files, interval, _log).run()
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    run_auto_reload()



# -*- coding: utf-8 -*-
import io
import os
import json
import zipfile
import contextlib
import tempfile
import shutil
import sys
import tarfile
import mimetypes
import subprocess
import click
import hashlib
from functools import partial


class AbortedError(Exception):
    pass


def get_content_type(name):
    return mimetypes.guess_type(name)[0] or "string"


def serialize_to_json(data):
    return json.dumps(data, indent=2, separators=(',', ': ')) + '\n'


class ImportStringError(ImportError):

    """Provides information about a failed :func:`import_string` attempt."""

    #: String in dotted notation that failed to be imported.
    import_name = None
    #: Wrapped exception.
    exception = None

    def __init__(self, import_name, exception):
        self.import_name = import_name
        self.exception = exception

        msg = (
            'import_string() failed for %r. Possible reasons are:\n\n'
            '- missing __init__.py in a package;\n'
            '- package or module path not included in sys.path;\n'
            '- duplicated package or module name taking precedence in '
            'sys.path;\n'
            '- missing module, class, function or variable;\n\n'
            'Debugged import:\n\n%s\n\n'
            'Original exception:\n\n%s: %s')

        name = ''
        tracked = []
        for part in import_name.replace(':', '.').split('.'):
            name += (name and '.') + part
            imported = import_string(name, silent=True)
            if imported:
                tracked.append((name, getattr(imported, '__file__', None)))
            else:
                track = ['- %r found in %r.' % (n, i) for n, i in tracked]
                track.append('- %r not found.' % name)
                msg = msg % (import_name, '\n'.join(track),
                             exception.__class__.__name__, str(exception))
                break

        ImportError.__init__(self, msg)

    def __repr__(self):
        return '<%s(%r, %r)>' % (self.__class__.__name__, self.import_name,
                                 self.exception)


def import_string(import_name, silent=False):
    import_name = str(import_name).replace(':', '.')
    try:
        try:
            __import__(import_name)
        except ImportError:
            if '.' not in import_name:
                raise
        else:
            return sys.modules[import_name]

        module_name, obj_name = import_name.rsplit('.', 1)
        try:
            module = __import__(module_name, None, None, [obj_name])
        except ImportError:
            # support importing modules not yet set up by the parent module
            # (or package for that matter)
            module = import_string(module_name)

        try:
            return getattr(module, obj_name)
        except AttributeError as e:
            raise ImportError(e)

    except ImportError as ex:
        if not silent:
            raise(
                ImportStringError,
                ImportStringError(import_name, ex),
                sys.exc_info()[2])


class OSUtils(object):
    ZIP_DEFLATED = zipfile.ZIP_DEFLATED

    def environ(self):
        return os.environ

    def open(self, filename, mode):
        return open(filename, mode)

    def open_zip(self, filename, mode, compression=ZIP_DEFLATED):
        return zipfile.ZipFile(filename, mode, compression=compression)

    def remove_file(self, filename):
        """Remove a file, noop if file does not exist."""
        # Unlike os.remove, if the file does not exist,
        # then this method does nothing.
        try:
            os.remove(filename)
        except OSError:
            pass

    def file_exists(self, filename):
        return os.path.isfile(filename)

    def get_file_contents(self, filename, binary=True, encoding='utf-8'):
        # It looks like the type definition for io.open is wrong.
        # the encoding arg is unicode, but the actual type is
        if binary:
            mode = 'rb'
            # In binary mode the encoding is not used and most be None.
            encoding = None
        else:
            mode = 'r'
        with io.open(filename, mode, encoding=encoding) as f:
            return f.read()

    def set_file_contents(self, filename, contents, binary=True):
        if binary:
            mode = 'wb'
        else:
            mode = 'w'
        with open(filename, mode) as f:
            f.write(contents)

    def extract_zipfile(self, zipfile_path, unpack_dir):
        with zipfile.ZipFile(zipfile_path, 'r') as z:
            z.extractall(unpack_dir)

    def extract_tarfile(self, tarfile_path, unpack_dir):
        with tarfile.open(tarfile_path, 'r:*') as tar:
            tar.extractall(unpack_dir)

    def directory_exists(self, path):
        return os.path.isdir(path)

    def get_directory_contents(self, path):
        return os.listdir(path)

    def makedirs(self, path):
        os.makedirs(path)

    def dirname(self, path):
        return os.path.dirname(path)

    def abspath(self, path):
        return os.path.abspath(path)

    def joinpath(self, *args):
        return os.path.join(*args)

    def walk(self, path):
        return os.walk(path)

    def copytree(self, source, destination):
        if not os.path.exists(destination):
            self.makedirs(destination)
        names = self.get_directory_contents(source)
        for name in names:
            new_source = os.path.join(source, name)
            new_destination = os.path.join(destination, name)
            if os.path.isdir(new_source):
                self.copytree(new_source, new_destination)
            else:
                shutil.copy2(new_source, new_destination)

    def rmtree(self, directory):
        shutil.rmtree(directory)

    def copy(self, source, destination):
        shutil.copy(source, destination)

    def move(self, source, destination):
        shutil.move(source, destination)

    def md5_file(self, file_name):
        with open(file_name, mode='rb') as f:
            d = hashlib.md5()
            for buf in iter(partial(f.read, 128), b''):
                d.update(buf)
        return d.hexdigest()

    @contextlib.contextmanager
    def tempdir(self):
        tempdir = tempfile.mkdtemp()
        try:
            yield tempdir
        finally:
            shutil.rmtree(tempdir)

    def popen(self, command, stdout=None, stderr=None, env=None):
        p = subprocess.Popen(command, stdout=stdout, stderr=stderr, env=env)
        return p

    def mtime(self, path):
        return os.stat(path).st_mtime

    @property
    def pipe(self):
        return subprocess.PIPE

    def recreate_dir(self, directory):
        if self.directory_exists(directory):
            self.rmtree(directory)
        self.makedirs(directory)


class UI(object):
    def __init__(self, out=None, err=None, confirm=None):
        # I tried using a more exact type for the 'confirm'
        # param, but mypy seems to miss the 'if confirm is None'
        # check and types _confirm as Union[..., None].
        # So for now, we're using Any for this type.
        if out is None:
            out = sys.stdout
        if err is None:
            err = sys.stderr
        if confirm is None:
            confirm = click.confirm
        self._out = out
        self._err = err
        self._confirm = confirm

    def write(self, msg):
        self._out.write(msg)

    def error(self, msg):
        self._err.write(msg)

    def confirm(self, msg, default=False, abort=False):
        try:
            return self._confirm(msg, default, abort)
        except click.Abort:
            raise AbortedError()



"""
Command line interface for alita.
"""
from __future__ import print_function, absolute_import

import os
import sys
import platform
import traceback
import click
import logging
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../")))

from alita import __version__
from alita.utils import OSUtils


def get_system_info():
    python_info = "python {}.{}.{}".format(sys.version_info[0],
                                           sys.version_info[1],
                                           sys.version_info[2])
    platform_system = platform.system().lower()
    platform_release = platform.release()
    platform_info = "{} {}".format(platform_system, platform_release)
    return "{}, {}".format(python_info, platform_info)


def config_logging(file_name=None):
    log_formatter = logging.Formatter("%(asctime)s: %(message)s")
    root_logger = logging.getLogger()

    if file_name:
        file_handler = logging.FileHandler(file_name)
        file_handler.setFormatter(log_formatter)
        root_logger.addHandler(file_handler)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)
    root_logger.addHandler(console_handler)


@click.group()
@click.version_option(version=__version__,
                      message='%(prog)s %(version)s, {}'
                      .format(get_system_info()))
@click.option('--project-dir',
              help='The project directory.  Defaults to CWD')
@click.option('--debug/--no-debug',
              default=False,
              help='Print debug logs to stderr.')
@click.pass_context
def cli(ctx, project_dir, debug=False):
    from alita.factory import CliFactory
    if project_dir is None:
        project_dir = os.getcwd()
    else:
        project_dir = os.path.abspath(project_dir)
    ctx.obj['project_dir'] = project_dir
    ctx.obj['debug'] = debug
    ctx.obj["os_utils"] = OSUtils()
    ctx.obj['factory'] = CliFactory()
    os.chdir(project_dir)
    config_logging()


@cli.command('run', short_help='Runs a development server.')
@click.option('--host', '-h', default='127.0.0.1',
              help='The interface to bind to.')
@click.option('--port', '-p', default=5000,
              help='The port to bind to.')
@click.option('--app', '-A', default='',
              help='The alita app module.')
@click.option('--auto-reload/--no-auto-reload', default=True,
              help='Enable or disable the reloader. By default the reloader '
              'is active if debug is enabled.')
@click.option('--debugger/--no-debugger', default=True,
              help='Enable or disable the debugger. By default the debugger '
              'is active if debug is enabled.')
@click.pass_context
def run(ctx, host, port, app, auto_reload, debugger):
    app = ctx.obj['factory'].load_app(app)
    app.run(host=host, port=port, auto_reload=auto_reload)


def main():
    try:
        return cli(obj={})
    except Exception as ex:
        print(ex)
        click.echo(traceback.format_exc(), err=True)
        return 2


if __name__ == '__main__':
    main()



from alita.app import Alita
from alita.worker import GunicornWorker
from alita.blueprints import Blueprint
from alita.response import *
from alita.templating import render_template, render_template_string
from alita.request import Request

__version__ = '0.2.1'



import os
import asyncio
import logging
import warnings
import itertools
import functools
import traceback
import alita.signals as signals
from alita.serve import *
from inspect import isawaitable
from alita.datastructures import ImmutableDict
from alita.config import Config, ConfigAttribute
from alita.factory import AppFactory
from alita.helpers import import_string, cached_property, method_dispatch
from alita.response import TextResponse, JsonResponse
from alita.exceptions import ServerError, BadRequest, WebSocketConnectionClosed
from alita.handler import IGNORE_EXCEPTIONS
from collections import UserDict
from alita.templating import Environment
from jinja2 import FileSystemLoader
from websockets import ConnectionClosed


class Alita(object):
    __charset__ = 'utf-8'
    config_class = Config
    _default_factory_class = AppFactory
    _view_middleware = []
    jinja_environment = Environment

    app_factory_class = None
    response_class = None
    exception_class = None

    key_prefix = ConfigAttribute('SESSION_KEY_PREFIX')
    session_use_signer = ConfigAttribute('SESSION_USE_SIGNER')
    session_engine = ConfigAttribute('SESSION_ENGINE')
    session_cookie_name = ConfigAttribute('SESSION_COOKIE_NAME')
    session_cookie_path = ConfigAttribute('SESSION_COOKIE_PATH')
    session_cookie_domain = ConfigAttribute('SESSION_COOKIE_DOMAIN')
    session_save_every_request = ConfigAttribute('SESSION_SAVE_EVERY_REQUEST')
    session_cookie_expire = ConfigAttribute('SESSION_COOKIE_EXPIRE')
    session_engine_config = ConfigAttribute('SESSION_ENGINE_CONFIG')
    session_table_name = ConfigAttribute('SESSION_TABLE_NAME')
    send_file_max_age = ConfigAttribute('SEND_FILE_MAX_AGE')
    max_content_length = ConfigAttribute('MAX_CONTENT_LENGTH')

    default_config = ImmutableDict({
        'DEBUG': False,
        'SECRET_KEY': None,
        'SESSION_COOKIE_EXPIRE': 30 * 24 * 60 * 60,
        'USE_X_SENDFILE': False,
        'SERVER_NAME': None,
        'APPLICATION_ROOT': '/',
        'SESSION_COOKIE_NAME': 'sessionid',
        'SESSION_COOKIE_DOMAIN': None,
        'SESSION_COOKIE_PATH': None,
        'SESSION_COOKIE_HTTPONLY': True,
        'SESSION_COOKIE_SECURE': False,
        'SESSION_COOKIE_SAMESITE': None,
        'SESSION_KEY_PREFIX': '',
        'SESSION_USE_SIGNER': False,
        'SESSION_REFRESH_EACH_REQUEST': True,
        'MAX_CONTENT_LENGTH': None,
        'SEND_FILE_MAX_AGE': 12 * 60 * 60,
        'TRAP_BAD_REQUEST_ERRORS': None,
        'TRAP_HTTP_EXCEPTIONS': False,
        'EXPLAIN_TEMPLATE_LOADING': False,
        'PREFERRED_URL_SCHEME': 'http',
        'JSON_AS_ASCII': True,
        'JSON_SORT_KEYS': True,
        'JSONIFY_PRETTYPRINT_REGULAR': False,
        'JSONIFY_MIMETYPE': 'application/json',
        'TEMPLATES_AUTO_RELOAD': False,
        'MAX_COOKIE_SIZE': 4093,
        'SESSION_SAVE_EVERY_REQUEST': False,
        'SESSION_EXPIRE_AT_BROWSER_CLOSE': False,
        'SESSION_MUST_SAVE': True,
        'SESSION_TABLE_NAME': 'session',
        "SESSION_ENGINE": None,
        'SESSION_ENGINE_CONFIG': None,
        'FORWARDED_FOR_HEADER': 'x-forwarded-for',
        'HTTP_HOST': 'http_host',
        'REAL_IP_HEADER': 'x-real-ip',
        'REMOTE_ADDR': 'remote_addr',
        'STRICT_SLASHES': True
    })

    def __init__(self, name=None, subdomain_matching=False, static_folder=None,
                 static_url_path=None, template_folder='templates'):
        self.name = name
        self.view_functions = {}
        self.view_functions_handlers = {}
        self.static_folder = static_folder
        self.static_url_path = static_url_path
        self.template_folder = template_folder
        self.subdomain_matching = subdomain_matching
        self.config = None
        self.is_running = False
        self.is_websocket = False
        self.make_config()

        self.before_request_funcs = {}
        self.after_request_funcs = {}
        self.error_handler_spec = {}
        self.blueprints = {}
        self._blueprint_order = []
        self.extensions = UserDict()
        self.logger = logging.getLogger(__name__)

        self.template_context_processors = {
            None: [self._default_template_ctx_processor]
        }
        self.app_factory_class = import_string(self.config.get(
            "APP_FACTORY_CLASS", self._default_factory_class))
        self.app_factory = self.app_factory_class(self)
        self.exception_handler = None
        self.static_handler = None
        self.router = None
        self.make_factory()
        self.websocket_tasks = set()

    def _get_debug(self):
        return self.config['DEBUG']

    def _set_debug(self, value):
        self.config['DEBUG'] = value
        self.jinja_env.auto_reload = self.templates_auto_reload()

    debug = property(_get_debug, _set_debug)
    del _get_debug, _set_debug

    def request_middleware(self, f):
        self.before_request_funcs.setdefault(None, []).append(f)
        return f

    def response_middleware(self, f):
        self.after_request_funcs.setdefault(None, []).append(f)
        return f

    def view_handler(self, f):
        self.view_functions_handlers.setdefault(None, []).append(f)
        return f

    def make_config(self):
        defaults = dict(self.default_config)
        self.config = self.config_class(defaults)

    def config_from_pyfile(self, filename, silent=False):
        self.config.from_pyfile(filename, silent)

    def config_from_jsonfile(self, filename, silent=False):
        self.config.from_json(filename, silent)

    def config_from_object(self, filename):
        self.config.from_object(filename)

    def make_factory(self):
        self.response_class = self.app_factory.create_base_response_class()
        self.exception_class = self.app_factory.create_base_exception_class()
        self.exception_handler = self.app_factory.create_exception_handler_object()
        self.router = self.app_factory.create_router_object()
        self.static_handler = self.app_factory.create_static_handler()

    def error_handler(self, code_or_exception):
        def decorator(func):
            self.register_error_handler(code_or_exception, func)
            return func
        return decorator

    def register_error_handler(self, code_or_exception, func):
        if isinstance(code_or_exception, int):
            self.exception_handler.add_status_handler(code_or_exception, func)
        elif issubclass(code_or_exception, Exception):
            self.exception_handler.add_exception_handler(code_or_exception, func)
        else:
            raise ValueError("decorator error handler code or exception "
                             "is invalid int or exception type!")

    @staticmethod
    def get_endpoint_from_view_func(view_func):
        return view_func.__name__

    def add_url_rule(self, view_func, rule, endpoint=None, methods=None, **options):
        if endpoint is None:
            endpoint = self.get_endpoint_from_view_func(view_func)
        if methods is None:
            methods = getattr(view_func, 'methods', None) or ('GET',)
        elif isinstance(methods, str):
            methods = [methods]
        assert isinstance(methods, (tuple, list))
        methods = set(item.upper() for item in methods)
        view_func = self.process_view_functions(view_func, endpoint, **options)
        self.check_view_functions(view_func, endpoint)
        self.router.add_route(rule, endpoint, view_func, methods)
        self.view_functions[endpoint] = view_func

    def check_view_functions(self, view_func, endpoint):
        old_func = self.view_functions.get(endpoint)
        if old_func is not None and old_func != view_func:
            raise AssertionError('View function mapping is overwriting an '
                                 'existing endpoint function: %s' % endpoint)
        if not asyncio.iscoroutinefunction(view_func):
            warnings.warn("View function [%s] should be async function" % view_func.__name__)

    def process_view_functions(self, view_func, endpoint=None, **options):
        bp = endpoint.rsplit('.', 1)[0] if endpoint and '.' in endpoint else None
        handlers = self.view_functions_handlers.get(None, [])
        if bp is not None and bp in self.view_functions_handlers:
            handlers = itertools.chain(handlers, self.view_functions_handlers[bp])
        for handler in handlers:
            view_func = handler(**options)(view_func)
        return view_func

    def route(self, rule, **options):
        def decorator(f):
            self.add_url_rule(f, rule,  **options)
            return f
        return decorator

    def run(self, extra_files=None, auto_reload=None, reload_interval=1, **kwargs):
        def inner(loop=None):
            server = Server(self, config=ServerConfig(loop=loop, **kwargs))
            server.run()

        def _log(type, message, *args, **kwargs):
            getattr(self.logger, type)(message.rstrip(), *args, **kwargs)

        if auto_reload and os.environ.get("SERVER_RUN_MAIN") != "true":
            run_auto_reload(extra_files, reload_interval, _log)
        else:
            inner()

    async def preprocess_request(self, request):
        bp = request.blueprint
        funcs = self.before_request_funcs.get(None, [])
        if bp is not None and bp in self.before_request_funcs:
            funcs = itertools.chain(funcs, self.before_request_funcs[bp])
        for func in funcs:
            rv = await self.get_awaitable_result(func, request)
            if rv is not None:
                return rv

    async def dispatch_request(self, request):
        if request.routing_exception is not None:
            raise request.routing_exception
        route_match = request.route_match
        return await self.get_awaitable_result(
            route_match.view_func,
            request, **route_match.path_params
        )

    @method_dispatch
    async def make_response(self, response):
        if callable(response):
            response = response()
        if isinstance(response, self.response_class):
            pass
        elif isinstance(response, self.exception_class):
            return response.get_response()
        else:
            raise ServerError(
                'The view function did not return a valid http response. The'
                ' function must returned a statement.'
            )
        return response

    @make_response.register(str)
    async def _(self, text):
        return TextResponse(text)

    @make_response.register(dict)
    async def _(self, json_value):
        return JsonResponse(json_value)

    @make_response.register(list)
    async def _(self, json_value):
        return JsonResponse(json_value)

    async def get_awaitable_result(self, func, *args, **kwargs):
        func_result = func(*args, **kwargs)
        if isawaitable(func_result):
            return await func_result
        else:
            return func_result

    async def process_response(self, request, response):
        bp = request.blueprint
        funcs = self.after_request_funcs.get(None, [])
        if bp is not None and bp in self.after_request_funcs:
            funcs = itertools.chain(funcs, self.after_request_funcs[bp])
        for func in funcs:
            response = await self.get_awaitable_result(func, request, response)
        return response

    async def finalize_response(self, response):
        return response

    async def finalize_request(self, request, response, from_error_handler=False):
        response = await self.make_response(response)
        try:
            response = await self.process_response(request, response)
            signals.request_finished.send(self, response=response)
        except Exception as ex:
            if not from_error_handler:
                raise ex
            self.logger.exception('Request finalizing failed with an '
                                  'error while handling an error')
        return response

    async def full_dispatch_request(self, request):
        try:
            signals.request_started.send(self)
            response = await self.preprocess_request(request)
            if response is None:
                response = await self.dispatch_request(request)
            response = await self.finalize_request(request, response)
            return await self.finalize_response(response)
        except Exception as ex:
            exception = await self.exception_handler.process_exception(request, ex)
            if issubclass(type(exception), BaseException):
                raise exception
            try:
                return await self.make_response(exception)
            except ServerError:
                message = "Caught handled exception: %s." % str(exception)
                self.logger.error(message)
                raise ServerError(message)

    async def create_request(self, environ):
        return self.app_factory.create_request_object(environ)

    async def __call__(self, environ, on_response):
        request, response = None, None
        try:
            request = await self.create_request(environ)
            response = await self.full_dispatch_request(request)
        except Exception as ex:
            try:
                exception = await self.exception_handler.process_exception(request, ex)
            except IGNORE_EXCEPTIONS:
                response = None
            else:
                if isinstance(exception, self.exception_class):
                    response = exception.get_response()
                elif isinstance(exception, self.response_class):
                    response = exception
                else:
                    response = None
        if not on_response:
            return response
        if response:
            await on_response(response)
        elif not self.is_websocket:
            message = "Caught handled exception, response object empty."
            self.logger.error(message)
            await on_response(self.exception_handler.process_exception(
                request, ServerError(message)).get_response())

    def log_exception(self, request, exc_info):
        try:
            self.logger.exception(traceback.format_exc())
            self.logger.error('Exception on %s [%s]' % (
                request.path,
                request.method
            ), exc_info=exc_info)
        except:
            pass

    def register_blueprint(self, blueprint, **options):
        if blueprint.name in self.blueprints:
            raise RuntimeError("blueprint %s has registered to app. "
                               "not allow repeat register." % blueprint.name)
        else:
            self.blueprints[blueprint.name] = blueprint
            self._blueprint_order.append(blueprint)
        blueprint.register(self, **options)

    def iter_blueprints(self):
        return iter(self._blueprint_order)

    def url_for(self, endpoint, **path_params):
        url_path = self.router.url_path_for(endpoint, **path_params)
        return url_path.make_url()

    def templates_auto_reload(self):
        r = self.config['TEMPLATES_AUTO_RELOAD']
        return r if r is not None else self.debug

    def create_jinja_environment(self):

        def select_jinja_autoescape(filename):
            if filename is None:
                return True
            return filename.endswith(('.html', '.htm', '.xml', '.xhtml'))

        options = dict(
            extensions=['jinja2.ext.autoescape', 'jinja2.ext.with_'],
            autoescape=select_jinja_autoescape,
            auto_reload=self.templates_auto_reload(),
            enable_async=True
        )
        rv = self.jinja_environment(self, **options)
        rv.globals.update(
            url_for=self.url_for,
            config=self.config,
        )
        return rv

    @cached_property
    def jinja_env(self):
        return self.create_jinja_environment()

    @cached_property
    def jinja_loader(self):
        if self.template_folder is not None:
            return FileSystemLoader(self.template_folder)

    def create_global_jinja_loader(self):
        return self.app_factory.create_jinja_loader()

    def add_template_filter(self, f, name=None):
        self.jinja_env.filters[name or f.__name__] = f

    def add_template_global(self, f, name=None):
        self.jinja_env.globals[name or f.__name__] = f

    def context_processor(self, f):
        assert asyncio.iscoroutinefunction(f)
        self.template_context_processors[None].append(f)
        return f

    async def _default_template_ctx_processor(self, request):
        return dict(request=request)

    def enable_websocket(self, enable=True):
        if not enable:
            for task in self.websocket_tasks:
                task.cancel()
        self.is_websocket = enable

    def add_websocket_handler(self, handler, rule, endpoint=None, subprotocols=None):
        """
        add a function to be registered as a websocket route
        :param handler: websocker handler view
        :param rule: path of the URL
        :param endpoint: view function endpoint
        :param subprotocols: websocket subprotocols
        :return: decorated function
        """
        self.enable_websocket()
        if not rule.startswith("/"):
            rule = "/" + rule

        @functools.wraps(handler)
        async def websocket_handler(request, *args, **kwargs):
            try:
                protocol = request.transport.get_protocol()
            except AttributeError:
                protocol = request.transport._protocol
            ws = await protocol.websocket_handshake(request, subprotocols)
            fut = asyncio.ensure_future(handler(request, ws, *args, **kwargs))
            self.websocket_tasks.add(fut)
            try:
                await fut
            except (asyncio.CancelledError, ConnectionClosed) as ex:
                if self.config.get("WRITE_WS_CONNECTION_CLOSED_LOG", False):
                    self.logger.exception(str(ex))
                raise WebSocketConnectionClosed
            finally:
                self.websocket_tasks.remove(fut)
            await ws.close()
            raise WebSocketConnectionClosed

        self.add_url_rule(
            websocket_handler, rule,
            methods=("GET", ),
            endpoint=endpoint
        )

    def websocket(self, rule, **options):
        """
        Decorate a function to be registered as a websocket route
        :param rule: path of the URL
        :param options: websocket function options
        :return: decorated function
        """
        def decorator(f):
            self.add_websocket_handler(f, rule,  **options)
            return f
        return decorator



from alita.serve.server import Server
from alita.serve.utils import STATUS_TEXT
from alita.serve.config import ServerConfig
from alita.serve.reloader import run_auto_reload


__all__ = [
    "Server",
    "STATUS_TEXT",
    "ServerConfig",
    "run_auto_reload"
]



# -*- coding: utf-8 -*-
import copy
from itertools import repeat

iteritems = lambda d, *args, **kwargs: iter(d.items(*args, **kwargs))


def is_immutable(self):
    raise TypeError('%r objects are immutable' % self.__class__.__name__)


def iter_multi_items(mapping):
    """Iterates over the items of a mapping yielding keys and values
    without dropping any from more complex structures.
    """
    if isinstance(mapping, MultiDict):
        for item in iteritems(mapping, multi=True):
            yield item
    elif isinstance(mapping, dict):
        for key, value in iteritems(mapping):
            if isinstance(value, (tuple, list)):
                for value in value:
                    yield key, value
            else:
                yield key, value
    else:
        for item in mapping:
            yield item


class _Missing(object):

    def __repr__(self):
        return 'no value'

    def __reduce__(self):
        return '_missing'


_missing = _Missing()


class ImmutableListMixin(object):
    _hash_cache = None

    def __hash__(self):
        if self._hash_cache is not None:
            return self._hash_cache
        rv = self._hash_cache = hash(tuple(self))
        return rv

    def __reduce_ex__(self, protocol):
        return type(self), (list(self),)

    def __delitem__(self, key):
        is_immutable(self)

    def __iadd__(self, other):
        is_immutable(self)
    __imul__ = __iadd__

    def __setitem__(self, key, value):
        is_immutable(self)

    def append(self, item):
        is_immutable(self)
    remove = append

    def extend(self, iterable):
        is_immutable(self)

    def insert(self, pos, value):
        is_immutable(self)

    def pop(self, index=-1):
        is_immutable(self)

    def reverse(self):
        is_immutable(self)

    def sort(self, cmp=None, key=None, reverse=None):
        is_immutable(self)


class ImmutableList(ImmutableListMixin, list):
    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            list.__repr__(self),
        )


class ImmutableDictMixin(object):
    _hash_cache = None

    @classmethod
    def fromkeys(cls, keys, value=None):
        instance = super(cls, cls).__new__(cls)
        instance.__init__(zip(keys, repeat(value)))
        return instance

    def __reduce_ex__(self, protocol):
        return type(self), (dict(self),)

    def _iter_hashitems(self):
        return iteritems(self)

    def __hash__(self):
        if self._hash_cache is not None:
            return self._hash_cache
        rv = self._hash_cache = hash(frozenset(self._iter_hashitems()))
        return rv

    def setdefault(self, key, default=None):
        is_immutable(self)

    def update(self, *args, **kwargs):
        is_immutable(self)

    def pop(self, key, default=None):
        is_immutable(self)

    def popitem(self):
        is_immutable(self)

    def __setitem__(self, key, value):
        is_immutable(self)

    def __delitem__(self, key):
        is_immutable(self)

    def clear(self):
        is_immutable(self)


class ImmutableMultiDictMixin(ImmutableDictMixin):
    def __reduce_ex__(self, protocol):
        return type(self), (list(iteritems(self, multi=True)),)

    def _iter_hashitems(self):
        return iteritems(self, multi=True)

    def add(self, key, value):
        is_immutable(self)

    def popitemlist(self):
        is_immutable(self)

    def poplist(self, key):
        is_immutable(self)

    def setlist(self, key, new_list):
        is_immutable(self)

    def setlistdefault(self, key, default_list=None):
        is_immutable(self)


class ImmutableDict(ImmutableDictMixin, dict):
    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            dict.__repr__(self),
        )

    def copy(self):
        return dict(self)

    def __copy__(self):
        return self


class UpdateDictMixin(object):

    """Makes dicts call `self.on_update` on modifications.

    .. versionadded:: 0.5

    :private:
    """

    on_update = None

    def calls_update(name):
        def oncall(self, *args, **kw):
            rv = getattr(super(UpdateDictMixin, self), name)(*args, **kw)
            if self.on_update is not None:
                self.on_update(self)
            return rv
        oncall.__name__ = name
        return oncall

    def setdefault(self, key, default=None):
        modified = key not in self
        rv = super(UpdateDictMixin, self).setdefault(key, default)
        if modified and self.on_update is not None:
            self.on_update(self)
        return rv

    def pop(self, key, default=_missing):
        modified = key in self
        if default is _missing:
            rv = super(UpdateDictMixin, self).pop(key)
        else:
            rv = super(UpdateDictMixin, self).pop(key, default)
        if modified and self.on_update is not None:
            self.on_update(self)
        return rv

    __setitem__ = calls_update('__setitem__')
    __delitem__ = calls_update('__delitem__')
    clear = calls_update('clear')
    popitem = calls_update('popitem')
    update = calls_update('update')
    del calls_update


class CallbackDict(UpdateDictMixin, dict):

    """A dict that calls a function passed every time something is changed.
    The function is passed the dict instance.
    """

    def __init__(self, initial=None, on_update=None):
        dict.__init__(self, initial or ())
        self.on_update = on_update

    def __repr__(self):
        return '<%s %s>' % (
            self.__class__.__name__,
            dict.__repr__(self)
        )


class TypeConversionDict(dict):
    def get(self, key, default=None, type=None):
        try:
            rv = self[key]
            if type is not None:
                rv = type(rv)
        except (KeyError, ValueError):
            rv = default
        return rv


class MultiDict(TypeConversionDict):

    """A :class:`MultiDict` is a dictionary subclass customized to deal with
    multiple values for the same key which is for example used by the parsing
    functions in the wrappers.  This is necessary because some HTML form
    elements pass multiple values for the same key.

    :class:`MultiDict` implements all standard dictionary methods.
    Internally, it saves all values for a key as a list, but the standard dict
    access methods will only return the first value for a key. If you want to
    gain access to the other values, too, you have to use the `list` methods as
    explained below.

    Basic Usage:

    >>> d = MultiDict([('a', 'b'), ('a', 'c')])
    >>> d
    MultiDict([('a', 'b'), ('a', 'c')])
    >>> d['a']
    'b'
    >>> d.getlist('a')
    ['b', 'c']
    >>> 'a' in d
    True

    It behaves like a normal dict thus all dict functions will only return the
    first value when multiple values for one key are found.

    From Werkzeug 0.3 onwards, the `KeyError` raised by this class is also a
    subclass of the :exc:`~exceptions.BadRequest` HTTP exception and will
    render a page for a ``400 BAD REQUEST`` if caught in a catch-all for HTTP
    exceptions.

    A :class:`MultiDict` can be constructed from an iterable of
    ``(key, value)`` tuples, a dict, a :class:`MultiDict` or from Werkzeug 0.2
    onwards some keyword parameters.

    :param mapping: the initial value for the :class:`MultiDict`.  Either a
                    regular dict, an iterable of ``(key, value)`` tuples
                    or `None`.
    """

    def __init__(self, mapping=None, key_error_exception=None):
        self.key_error_exception = key_error_exception
        if isinstance(mapping, MultiDict):
            dict.__init__(self, ((k, l[:]) for k, l in mapping.lists()))
        elif isinstance(mapping, dict):
            tmp = {}
            for key, value in iteritems(mapping):
                if isinstance(value, (tuple, list)):
                    if len(value) == 0:
                        continue
                    value = list(value)
                else:
                    value = [value]
                tmp[key] = value
            dict.__init__(self, tmp)
        else:
            tmp = {}
            for key, value in mapping or ():
                tmp.setdefault(key, []).append(value)
            dict.__init__(self, tmp)

    def __getstate__(self):
        return dict(self.lists())

    def __setstate__(self, value):
        dict.clear(self)
        dict.update(self, value)

    def __getitem__(self, key):
        """Return the first data value for this key;
        raises KeyError if not found.

        :param key: The key to be looked up.
        :raise KeyError: if the key does not exist.
        """
        if key in self:
            lst = dict.__getitem__(self, key)
            if len(lst) > 0:
                return lst[0]
        if self.key_error_exception:
            raise self.key_error_exception(key)

    def __setitem__(self, key, value):
        """Like :meth:`add` but removes an existing key first.

        :param key: the key for the value.
        :param value: the value to set.
        """
        dict.__setitem__(self, key, [value])

    def add(self, key, value):
        """Adds a new value for the key.

        .. versionadded:: 0.6

        :param key: the key for the value.
        :param value: the value to add.
        """
        dict.setdefault(self, key, []).append(value)

    def getlist(self, key, type=None):
        """Return the list of items for a given key. If that key is not in the
        `MultiDict`, the return value will be an empty list.  Just as `get`
        `getlist` accepts a `type` parameter.  All items will be converted
        with the callable defined there.

        :param key: The key to be looked up.
        :param type: A callable that is used to cast the value in the
                     :class:`MultiDict`.  If a :exc:`ValueError` is raised
                     by this callable the value will be removed from the list.
        :return: a :class:`list` of all the values for the key.
        """
        try:
            rv = dict.__getitem__(self, key)
        except KeyError:
            return []
        if type is None:
            return list(rv)
        result = []
        for item in rv:
            try:
                result.append(type(item))
            except ValueError:
                pass
        return result

    def setlist(self, key, new_list):
        """Remove the old values for a key and add new ones.  Note that the list
        you pass the values in will be shallow-copied before it is inserted in
        the dictionary.

        >>> d = MultiDict()
        >>> d.setlist('foo', ['1', '2'])
        >>> d['foo']
        '1'
        >>> d.getlist('foo')
        ['1', '2']

        :param key: The key for which the values are set.
        :param new_list: An iterable with the new values for the key.  Old values
                         are removed first.
        """
        dict.__setitem__(self, key, list(new_list))

    def setdefault(self, key, default=None):
        """Returns the value for the key if it is in the dict, otherwise it
        returns `default` and sets that value for `key`.

        :param key: The key to be looked up.
        :param default: The default value to be returned if the key is not
                        in the dict.  If not further specified it's `None`.
        """
        if key not in self:
            self[key] = default
        else:
            default = self[key]
        return default

    def setlistdefault(self, key, default_list=None):
        """Like `setdefault` but sets multiple values.  The list returned
        is not a copy, but the list that is actually used internally.  This
        means that you can put new values into the dict by appending items
        to the list:

        >>> d = MultiDict({"foo": 1})
        >>> d.setlistdefault("foo").extend([2, 3])
        >>> d.getlist("foo")
        [1, 2, 3]

        :param key: The key to be looked up.
        :param default: An iterable of default values.  It is either copied
                        (in case it was a list) or converted into a list
                        before returned.
        :return: a :class:`list`
        """
        if key not in self:
            default_list = list(default_list or ())
            dict.__setitem__(self, key, default_list)
        else:
            default_list = dict.__getitem__(self, key)
        return default_list

    def items(self, multi=False):
        """Return an iterator of ``(key, value)`` pairs.

        :param multi: If set to `True` the iterator returned will have a pair
                      for each value of each key.  Otherwise it will only
                      contain pairs for the first value of each key.
        """

        for key, values in iteritems(dict, self):
            if multi:
                for value in values:
                    yield key, value
            else:
                yield key, values[0]

    def lists(self):
        """Return a list of ``(key, values)`` pairs, where values is the list
        of all values associated with the key."""

        for key, values in iteritems(dict, self):
            yield key, list(values)

    def values(self):
        """Returns an iterator of the first value on every key's value list."""
        for values in self.values():
            yield values[0]

    def copy(self):
        """Return a shallow copy of this object."""
        return self.__class__(self)

    def deepcopy(self, memo=None):
        """Return a deep copy of this object."""
        return self.__class__(copy.deepcopy(self.to_dict(flat=False), memo))

    def to_dict(self, flat=True):
        """Return the contents as regular dict.  If `flat` is `True` the
        returned dict will only have the first item present, if `flat` is
        `False` all values will be returned as lists.

        :param flat: If set to `False` the dict returned will have lists
                     with all the values in it.  Otherwise it will only
                     contain the first value for each key.
        :return: a :class:`dict`
        """
        if flat:
            return dict(iteritems(self))
        return dict(self.lists())

    def update(self, other_dict):
        for key, value in iter_multi_items(other_dict):
            MultiDict.add(self, key, value)

    def pop(self, key, default=_missing):
        """Pop the first item for a list on the dict.  Afterwards the
        key is removed from the dict, so additional values are discarded:

        >>> d = MultiDict({"foo": [1, 2, 3]})
        >>> d.pop("foo")
        1
        >>> "foo" in d
        False

        :param key: the key to pop.
        :param default: if provided the value to return if the key was
                        not in the dictionary.
        """
        try:
            lst = dict.pop(self, key)

            if len(lst) == 0:
                if self.key_error_exception:
                    raise self.key_error_exception(key)

            return lst[0]
        except KeyError as e:
            if default is not _missing:
                return default
            if self.key_error_exception:
                raise self.key_error_exception(str(e))

    def popitem(self):
        """Pop an item from the dict."""
        try:
            item = dict.popitem(self)

            if len(item[1]) == 0:
                if self.key_error_exception:
                    raise self.key_error_exception()

            return (item[0], item[1][0])
        except KeyError as e:
            if self.key_error_exception:
                raise self.key_error_exception(str(e))

    def poplist(self, key):
        """Pop the list for a key from the dict.  If the key is not in the dict
        an empty list is returned.

        .. versionchanged:: 0.5
           If the key does no longer exist a list is returned instead of
           raising an error.
        """
        return dict.pop(self, key, [])

    def popitemlist(self):
        """Pop a ``(key, list)`` tuple from the dict."""
        try:
            return dict.popitem(self)
        except KeyError as e:
            if self.key_error_exception:
                raise self.key_error_exception(str(e))

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.deepcopy(memo=memo)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, list(iteritems(self, multi=True)))
